"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper

from dune.perftool.generation import (domain,
                                      globalarg,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      valuearg,
                                      )

from dune.perftool.pdelab.basis import lfs_iname, lfs_iname_bound


@iname
def index_sum_iname(i):
    from dune.perftool.pdelab import name_index
    return name_index(i)


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper):
    def __init__(self):
        super(UFL2LoopyVisitor, self).__init__()

    def coefficient(self, o):
        # All trial functions should already be handled
        assert o.count() != 0
        # TODO implement non-trialfunction coefficients
        raise NotImplementedError

    def index_sum(self, o):
        from dune.perftool.ufl.shape import determine_shape

        # Define an iname for each of the indices in the multiindex
        for i in o.ufl_operands[1].indices():
            shape = determine_shape(o.ufl_operands[0], i)
            index_sum_iname(i)
            from dune.perftool.pdelab import name_index
            domain(name_index(i), shape)

        # Now continue processing the expression
        return self.call(o.ufl_operands[0])

    # TODO use multiple inheritance and have a geometry transformer in the pdelab subpackage
    def facet_area(self, o):
        from pymbolic.primitives import Variable
        from dune.perftool.pdelab.geometry import name_facetarea
        return Variable(name_facetarea())


class _Counter:
    counter = 0


def get_count():
    c = _Counter.counter
    _Counter.counter = c + 1
    return c


def transform_accumulation_term(term):
    from dune.perftool.ufl.transformations.replace import ReplaceExpression
    from pymbolic.primitives import Variable

    # We always have a quadrature loop
    from dune.perftool.pdelab.quadrature import quadrature_iname
    quadrature_iname()

    # Get the pymbolic expression needed for this accumulation term.
    # This includes filling the cache with all sorts of necessary preambles!
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    test_ma = extract_modified_arguments(term, trialfunction=False, testfunction=True)
    trial_ma = extract_modified_arguments(term, trialfunction=True, testfunction=False)

    rmap = {}
    for ma in test_ma:
        from dune.perftool.pdelab.argument import pymbolic_testfunction
        rmap[ma.expr] = pymbolic_testfunction(ma)
    for ma in trial_ma:
        from dune.perftool.pdelab.argument import pymbolic_trialfunction
        rmap[ma.expr] = pymbolic_trialfunction(ma)

    # Get the transformer!
    ufl2l_mf = UFL2LoopyVisitor()
    re_mf = ReplaceExpression(replacemap=rmap, otherwise=ufl2l_mf)
    ufl2l_mf.call = re_mf.__call__

    pymbolic_expr = re_mf(term)

    # Now simplify the expression
    # TODO: Add a switch to disable/configure this.
#     from dune.perftool.pymbolic.simplify import simplify_pymbolic_expression
#     pymbolic_expr = simplify_pymbolic_expression(pymbolic_expr)

    # Define a temporary variable for this expression
    expr_tv_name = "expr_" + str(get_count()).zfill(4)
    expr_tv = temporary_variable(expr_tv_name)
    from pymbolic.primitives import Variable
    instruction(assignee=Variable(expr_tv_name), expression=pymbolic_expr)

    # The data that is used to collect the arguments for the accumulate function
    accumargs = []

    # Generate the code for the modified arguments:
    for arg in test_ma:
        from dune.perftool.pdelab.argument import name_argumentspace, pymbolic_argument
        accumargs.append(name_argumentspace(arg))
        accumargs.append(lfs_iname(arg.argexpr.element(), argcount=arg.argexpr.count()))

    from dune.perftool.pdelab.argument import name_residual
    residual = name_residual()

    # The residual/the jacobian should be represented through a loopy global argument
    from dune.perftool.ufl.rank import ufl_rank
    globalarg(residual, shape=tuple(lfs_iname_bound(i) for i in range(ufl_rank(term))))

    from dune.perftool.generation import retrieve_cache_items
    inames = retrieve_cache_items("iname")

    from dune.perftool.pdelab.quadrature import name_factor
    instruction(inames=inames,
                code="{}.accumulate({}, {}*{})".format(residual,
                                                       ", ".join(accumargs),
                                                       expr_tv_name,
                                                       name_factor()
                                                       ),
                assignees=residual,
                )
